open Format

(* command line option *)

let short_usage = "maxcomp v1.1\nUsage: maxcomp <options> <file>\n"
let filenames = ref []

let d = ref 2
let k = ref 2
let kk = ref 1

let order = ref Maxtrs.lpo
       
(* main *)
let read_m s =
  match [ int_of_string x | x <- Str.split (Str.regexp ",") s ] with
  | [d; k] -> Mi.order ~d ~k
  | _ -> failwith "Invalid argument of -m"

let options = Arg.align 
  [("-lpo", Arg.Unit (fun s -> order := Maxtrs.lpo),
    " use LPO");
   ("-mpo", Arg.Unit (fun s -> order := Maxtrs.mpo),
    " use MPO");
   ("-kbo", Arg.Unit (fun s -> order := Maxtrs.kbo),
    " use KBO");
   ("-mi", Arg.String (fun s -> order := read_m s),
    "<d,n> use d*d matrix interpretaions with coefficients <= n");
   ("-K", Arg.Int (fun n -> kk := n),
    "<n> compute n maximal terminating TRSs")
 ]

let print_trs ppf rules = 
  fprintf ppf "(VAR %s)@.(RULES@.%a@.)@."
    (String.concat " " (Rules.variables rules))
    Rules.print rules

let () =
  Arg.parse options 
   (fun x -> filenames := !filenames @ [x]) short_usage;
  match !filenames with
  | [f] ->  
      let es = Rules.rpl_spcl_char (Read.read_trs f) in
      begin try
	let trs, _ = Maxcomp.maxcomp ~max:(Maxtrs.max ~order:!order) ~kk:!kk es in
	printf "YES@.%a@." print_trs trs;
      with 
        e -> printf "MAYBE@."; raise e
      end
  | _ -> eprintf "%s%!" short_usage; exit 1
