open Overlap
open Variant
open Rewriting

type result = 
  | TRS of Rules.t  (* Complete TRS *)
  | SC of Rules.t   (* S(C) *)

type max = int -> Rules.t -> Rules.t list

type env = 
  { fs : string list; max : max; ee : Rules.t;
    kk : int; d : int; n : int; max_found : int } 

let union_es es1 es2 = unique ~eq:eq_variant (es1 @ es2)

let sort_by f xs = List.sort (fun x y -> f x - f y) xs

(* Size filtering *)
let filter env cc = 
  Listx.take_at_most env.n 
    (sort_by Rule.size [ e | e <- cc; Rule.size e < env.d])

exception Fail of string

(* Non-trivial part of { nf(s) = nf(t) | s = t in E cup CP(R) } *)
let unjoined ee trs =
  let trs_r = right_reduce trs in  (* optimization *)
  rename_rules 
    [ s, t | u, v <- ee @ cp trs_r; 
	     s, t <- [nf trs_r u, nf trs_r v]; s <> t ]

(* The function returns either a complete TRS for E or S(C). *)
let rec trs_or_sc env ?(acc = []) = function
  | [] -> SC acc 
  | trs :: trss -> 
      match unjoined env.ee trs with
      | [] -> TRS trs
      | sc -> trs_or_sc env ~acc:(union_es acc (filter env sc)) trss 

let rec phi env cc =
  let trss = env.max env.kk cc in (* R(C) *)
  if List.length trss = env.max_found then 
    raise (Fail "I'm giving up\n")
  else
    match trs_or_sc env trss with
    | TRS trs                    -> trs, env.kk
    | SC sc when eq_subset sc cc -> phi { env with kk = env.kk + 1; max_found = List.length trss } cc
    | SC sc                      -> phi { env with max_found = 0} (union_es cc sc)

let maxcomp ~max ~kk ee =
  let env = 
    { fs = Rules.functions ee; ee = ee; max = max; 
      kk = kk; n = 7; d = 20; max_found = 0 } in
    let trs, kk = phi env ee in
     (unique ~eq:Rule.variant (reduce trs)), kk
