open Term
open Yices
open Arith

let rec emb_geq s t = 
  match s, t with 
  | V x, V y when x = y -> true
  | F (f, ss), (F (g, ts) as t) when f = g ->
      List.exists (fun si -> emb_geq si t) ss || 
      List.for_all2 (fun si ti -> emb_geq si ti) ss ts
  | F (f, ss), t -> 
      List.exists (fun si -> emb_geq si t) ss
  | _ -> false

let emb_gt s t = s <> t && emb_geq s t


let rec one ctx bs =
  match bs with
  | [] -> mk_false ctx
  | bi::bss ->
      mk_and ctx [|
      mk_or ctx [| mk_not ctx bi; zero ctx bss |];
      mk_or ctx [| bi; one ctx bss |] |]
and zero ctx bs = 
  match bs with
  | [] -> mk_true ctx
  | bi::bss ->
      mk_and ctx [| mk_not ctx bi; zero ctx bss |]

let c_yijei ctx yij ei m =
  let yijnj = fun lss j -> [ List.nth ls j | ls <- lss ] in
  let c_yij = 
    conj ctx [ one ctx (yijnj yij j) | j <- Listx.interval 0 (m - 1) ] in
  let c_ei =  mk_and ctx (Array.of_list [ mk_or ctx [|mk_not ctx ei_; one ctx yxj|] |
                                         ei_,yxj <- (List.combine ei yij) ]) in
  mk_and ctx [|c_yij; c_ei|]

let rec yrpo ctx a s t =
  match s with
  | F(f, ss) ->
      mk_or ctx (Array.of_list (yrpo2 ctx a s t :: 
 	[ mk_or ctx [|yrpo ctx a si t; yrpo_eq ctx a si t|] | si <- ss] ))
  | _ -> mk_false ctx
and yrpo2 ctx a s t =
 match s,t with
 | F(f, ss), F(g, tt) ->
   let c1 = mk_and ctx (Array.of_list (mk_true ctx :: [ yrpo ctx a s tj | tj <- tt ])) in
   let c2 = mk_or ctx 
          [| mk_gt ctx (List.assoc f a) (List.assoc g a);
             mk_and ctx [|mk_eq ctx (List.assoc f a) (List.assoc g a); ymul ctx a ss tt |]  
           |] in
   mk_and ctx [|c1;c2|]
 | _ -> mk_false ctx
and yrpo_eq ctx a s t =
 if s = t then mk_true ctx
 else 
  match s, t with
   | F(f, ss), F(g, tt) -> 
        mk_and ctx [| mk_eq ctx (List.assoc f a) (List.assoc g a); ymul_eq ctx a ss tt |]
   | _ -> mk_false ctx
and ymul_geq ctx a tpls =
 let cij si ei yij tj =
   let c_eq = mk_or ctx [| mk_not ctx ei; yrpo_eq ctx a si tj|] in
   let c_gr = mk_or ctx [| ei; yrpo ctx a si tj |] in
    mk_or ctx [| mk_not ctx yij; mk_and ctx [|c_eq; c_gr|] |]
 in
   mk_and ctx (Array.of_list [ cij si ei yij tj | si,ei,ls <- tpls; yij, tj <- ls ])
and ymul ctx a ss tt = 
 match ss, tt with
   | [], _ -> mk_false ctx
   | _ , [] -> mk_true ctx
   | _ , _ -> 
    let tpls = [ si, mk_fresh_bool_var ctx, 
                  [ mk_fresh_bool_var ctx, tj | tj <- tt ] 
                | si <- ss ] 
    in
    let ei = List.map (fun (x,y,z) -> y) tpls in
    let frst = fun (x,y) -> x in
    let yij = List.map (fun (x,y,z) -> (List.map frst z)) tpls in
    let m = List.length tt in 
    let c_not_ei = mk_not ctx (mk_and ctx (Array.of_list ei)) in
    let c_yijei = c_yijei ctx yij ei m in
     mk_and ctx [| ymul_geq ctx a tpls; c_not_ei; c_yijei |]
and ymul_eq ctx a ss tt =
 match ss, tt with
   | [], [] -> mk_true ctx
   | _ , [] -> mk_false ctx
   | [],  _ -> mk_false ctx
   | _ , _ -> 
    let tpls = [ si, mk_fresh_bool_var ctx, 
                  [ mk_fresh_bool_var ctx, tj | tj <- tt ] 
                | si <- ss ] 
    in
    let ei = List.map (fun (x,y,z) -> y) tpls in
    let frst = fun (x,y) -> x in
    let yij = List.map (fun (x,y,z) -> (List.map frst z)) tpls in
    let m = List.length tt in 
    let c_ei = mk_and ctx (Array.of_list ei) in
    let c_yijei = c_yijei ctx yij ei m in
     mk_and ctx [| ymul_geq ctx a tpls; c_ei; c_yijei |]


(* lpo for yices *)

let rec ylpo ctx a s t =
  (* optimization *)
  if emb_gt s t then mk_true ctx else
  if not (Listset.subset (variables t) (Term.variables s)) || 
     emb_geq t s then mk_false ctx else 
  match s with
  | V _-> mk_false ctx
  | F (f, ss)  ->
      let y = 
        match t with
	| V x -> if List.mem x (variables s) then mk_true ctx else mk_false ctx
	| F (g, ts) when f = g ->
            mk_and ctx (Array.of_list (ylex ctx a ss ts :: [ ylpo ctx a s ti | ti <- ts ])
                       )
	| F (g, ts) ->
            mk_and ctx (Array.of_list (mk_gt ctx (List.assoc f a) (List.assoc g a) :: 
				       [ ylpo ctx a s ti | ti <- ts ]))
      in
      mk_or ctx (Array.of_list (y :: [ ylpo_geq ctx a si t | si <- ss ]))
and ylpo_geq ctx a s t = 
  if s = t then mk_true ctx else ylpo ctx a s t
and ylex ctx a l1 l2 =
  match l1, l2 with
  | s :: ss, t :: ts when s = t -> ylex ctx a ss ts
  | s :: ss, t :: ts -> ylpo ctx a s t
  | [], [] -> mk_false ctx
  | _ -> mk_true ctx


(* kbo for yices *)

type kbo_wf = (Yices.expr * Yices.expr * string) list
type kbo_wf_w0 = kbo_wf * Yices.expr
type kbo_wf_wc_w1_w0 = kbo_wf * kbo_wf * kbo_wf * Yices.expr

(* 
(wf,pf,f) for all f \ín F
(wc,pc,c) for all c \in F
(wf,pc,c) for all f \in F with one argument
w0
*)

let adm_smt ctx (wfs,wcs,w1s,w0) =
  let zero = mk_num ctx 0 in
  let c_w0 = mk_gt ctx w0 zero in
  let ensure_nat = mk_and ctx 
      (Array.of_list (List.flatten 
			[ [mk_ge ctx w zero; mk_ge ctx p zero] | w,p,f <- wfs])) in
  let c_wcs = 
    (if wcs = [] then mk_true ctx 
    else
      mk_and ctx (Array.of_list [ mk_ge ctx wc w0 | wc,_,_ <- wcs ] ))
  in
  let pf_geq_pg pf = mk_and ctx (Array.of_list [ mk_ge ctx pf pg | _,pg,_ <- wfs ]) in
  let not_wf_eq_0 wf = mk_not ctx (mk_eq ctx wf zero) in
  let c_w1s = 
    (if w1s = [] then mk_true ctx 
    else
      mk_and ctx  (Array.of_list 
		     [ mk_or ctx (Array.of_list [ not_wf_eq_0 wf ; pf_geq_pg pf]) | wf,pf,_ <- w1s ]
		  ) 
    )
  in
  mk_and ctx [|ensure_nat;c_w0;c_wcs;c_w1s|]
    
let rec kbo_weight ctx s (wfs,w0) =
  match s with
  | V _ -> w0 
  | F(f,ss) ->
      let wf,_,_ = List.find (fun (w,p,g) -> g = f) wfs in
      if ss = [] then wf
      else
	mk_sum ctx (Array.of_list 
		      (wf :: [kbo_weight ctx si (wfs,w0) | si <- ss ]))


let kbo_var_cond s t =
 let vars = Rule.variables (s,t) in
 if List.exists (fun v -> 
   (Term.count_variable v s) 
     < (Term.count_variable v t)) 
     vars
 then false else true


let rec first_diff ss tt =
  match ss,tt with
  | [], _ -> []
  | _, [] -> []
  | s::sss, t::ttt -> 
      if s <> t then [(s,t)]
      else first_diff sss ttt

	  


exception Enc_kbo_error

(* checks wether s = f(f(...(x))) and t = x *)
(*   returns true if s=x and t=x !
   does not ensure t = Var! *)
let rec kbo_b1 s t =
  if s = t then true
  else
    match s with
    | F(f,ss) when List.length ss = 1 -> kbo_b1 (List.nth ss 0) t
    | _ -> false

let rec ykbo ctx (wfs,w0) s t = 
  match s with
  | V _ -> mk_false ctx
  | F(f,ss) -> 
      if (s = t || not (kbo_var_cond s t)) 
      then mk_false ctx
      else 
	let weight_s = kbo_weight ctx s (wfs,w0) in
	let weight_t = kbo_weight ctx t (wfs,w0) in
	mk_or ctx
	  [| mk_gt ctx weight_s weight_t;	 
             mk_and ctx [| mk_eq ctx weight_s weight_t; 
         		   ykbo1 ctx (wfs,w0) s t |]
	   |]
and ykbo1 ctx (wfs,w0) s t =
  match s, t with
  | F(f,ss), V _ when List.length ss = 1 ->
      if kbo_b1 s t then mk_true ctx
      else mk_false ctx
  | F(f,ss), F(g,tt) ->
      (
       let _,pf,_ = List.find (fun (w,p,h) -> h = f) wfs in
       let _,pg,_ = List.find (fun (w,p,h) -> h = g) wfs in 
       let siti = first_diff ss tt in       
       if siti = [] then mk_gt ctx pf pg
       else 
	 let si, ti = List.nth siti 0 in
	 mk_or ctx [|mk_gt ctx pf pg;
		     mk_and ctx [| mk_eq ctx pf pg; ykbo ctx (wfs,w0) si ti|]
		   |]
      )
  | _, _ -> mk_false ctx     
