open Term
open Subst
open Format

type t = Rule.t * int list * Rule.t * Subst.t

let count = ref 0

let fresh () = incr count; sprintf "__overlap_x%d" !count

let rename (l, r) =  
  let s = [ x, V (fresh ()) | x <- variables l ] in
  (substitute s l, substitute s r)
      
let overlap_aux rule1 rule2 =
  let l1, r1 = rename rule1
  and l2, r2 = rename rule2 in
  [ (l1, r1), p, (l2, r2), mgu (subterm_at p l2) l1
  | p <- function_positions l2; 
    unifiable (subterm_at p l2) l1 && 
    (p <> [] || not (Rule.variant (l1, r1) (l2, r2))) ] 

let overlap2 rules1 rules2 = 
  Listset.unique
    [ x | rule1 <- rules1; rule2 <- rules2; 
          x <- overlap_aux rule1 rule2 ]

let overlap rules = overlap2 rules rules

let cp_of_overlap ((l1, r1), p, (l2, r2), mu) =
  (Term.substitute mu (replace l2 r1 p), Term.substitute mu r2)

let cp2 rules1 rules2 = 
  [ cp_of_overlap o | o <- overlap2 rules1 rules2 ]

let cp rules = cp2 rules rules

let is_overlap rule1 p rule2 =
  let l1, r1 = rename rule1 
  and l2, r2 = rename rule2 in
  unifiable l1 (subterm_at p l2) && 
  (p <> [] || not (Rule.variant (l1, r1) (l2, r2)))

let overlapping_aux (((l1, r1) as rule1), ((l2, r2) as rule2)) =
  List.exists 
    (fun p -> is_overlap rule1 p rule2)
    (function_positions l2)

let overlapping rules = 
  List.exists overlapping_aux [ x,y | x <- rules; y <- rules ]

let non_overlapping rules = not (overlapping rules)
