open Term

let rec rewrite_at_root t = function
  | [] -> false, t
  | (l, r) :: rules ->
      try
	true, substitute (Subst.pattern_match l t) r
      with
	Subst.Not_matched -> rewrite_at_root t rules

let rec rewrite_aux rules = function
  | V _ as t -> false, t
  | F (f, ts) ->
      let l = [ rewrite_aux rules ti | ti <- ts ] in
      let b, u = rewrite_at_root (F (f, [ t | _, t <- l ])) rules in
      b || List.exists (fun (b, _) -> b) l, u

let rec nf rules t =
  let b, u = rewrite_aux rules t in
  if b then nf rules u else t

let rec replace_at i ts u =
  match i, ts with
  | _, [] -> invalid_arg "Rewriting.replace_at"
  | 0, _ :: l -> u :: l
  | i, x :: l -> x :: replace_at (i - 1) l u

let rec rewrite rules = function
  | V _ -> []
  | F (f, ts) as t ->
      Listset.union 
	(Listset.unique
         [ substitute (Subst.pattern_match l t) r
         | l, r <- rules; Subst.is_instance_of t l ])
	[ F (f, replace_at i ts u) 
        | i, ti <- Listx.ix ts; u <- rewrite rules ti ]
